"""Decorators for internal use in clients module."""
import functools
from typing import Callable, TypeVar


T = TypeVar("T")


def check_conn(func: Callable[..., T]) -> Callable[..., T]:  # noqa
    """Check that the connection to client is done otherwise raise an exception."""

    @functools.wraps(func)
    def _wrapper(*args, **kwargs) -> T:
        # Instance is passed as first positional argument
        inst = args[0]

        if hasattr(inst, "closed"):
            if inst.closed:
                raise ConnectionError("The connection is closed")
        else:
            raise AttributeError("Missing instance closed attribute")

        return func(*args, **kwargs)

    return _wrapper



"""Define default copier."""
from typing import cast

from tentaclio.protocols import Reader, Writer
from tentaclio.streams.api import open
from tentaclio.urls import URL


__all__ = ["DefaultCopier"]


class DefaultCopier:
    """Copier for those schemas that don't have an specialised implementation."""

    def copy(self, source: URL, dest: URL):
        """Copy the contents of the source url into the dest url."""
        with open(str(source), mode="rb") as reader, open(str(dest), mode="wb") as writer:
            cast(Writer, writer).write(cast(Reader, reader).read())



"""Base handler."""
import io
import logging
from typing import Callable

from tentaclio.protocols import ReaderClosable, WriterClosable
from tentaclio.streams import base_stream
from tentaclio.urls import URL


logger = logging.getLogger(__name__)

StreamerFactory = Callable[..., base_stream.StreamerContextManager]

__all__ = ["StreamURLHandler"]


def _is_bytes_mode(mode: str) -> bool:
    if "b" in str(mode):
        return True
    return False


class StreamURLHandler:
    """Handler for opening writers and readers ."""

    client: StreamerFactory

    def __init__(self, client_factory: StreamerFactory):
        """Create a handler using  a stream client factory to instantiate the underlying client."""
        self.client_factory = client_factory

    def open_reader_for(self, url: URL, mode: str, extras: dict) -> ReaderClosable:
        """Open an stream client for reading."""
        client = self.client_factory(url, **extras)

        if _is_bytes_mode(mode):
            return base_stream.StreamerReader(client, io.BytesIO())
        return base_stream.StringToBytesClientReader(client)

    def open_writer_for(self, url: URL, mode: str, extras: dict) -> WriterClosable:
        """Open an stream client writing."""
        client = self.client_factory(url, **extras)

        if _is_bytes_mode(mode):
            return base_stream.StreamerWriter(client, io.BytesIO())
        return base_stream.StringToBytesClientWriter(client)



"""Stream handler registry to open readers and writers to urls."""
import abc
import logging
from typing import Callable, ClassVar, ContextManager, Iterable, Protocol

from tentaclio import clients
from tentaclio.registry import URLHandlerRegistry


logger = logging.getLogger(__name__)

__all__ = ["DB_REGISTRY", "Db"]


class Db(ContextManager, Protocol):
    """Interface for query-based connections."""

    # Query methods:

    @abc.abstractmethod
    def execute(self, sql_query: str, **params) -> None:
        """Execute a query against the underlying client."""
        ...

    @abc.abstractmethod
    def query(self, sql_query: str, **params) -> Iterable:
        """Perform the query and return an iterable of the results."""
        ...


DbFactory = Callable[..., Db]


class DbRegistry(URLHandlerRegistry[DbFactory]):
    """Registry for databases."""

    def get_handler(self, scheme: str) -> DbFactory:
        """Get the handler for the give scheme. Raise an URLError if no handler is registred."""
        if scheme not in self.registry:
            logger.info("trying to return the default SQLA client")
            return clients.SQLAlchemyClient

        return self.registry[scheme]


class _DbRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[DbRegistry] = DbRegistry()


DB_REGISTRY = _DbRegistryHolder().instance



"""FTP ans SFTP stream clients."""
import ftplib
import io
import logging
import stat
from typing import Iterable, Optional, Union

import pysftp

from tentaclio import fs, protocols, urls

from . import base_client, decorators, exceptions


logger = logging.getLogger(__name__)

__all__ = ["FTPClient", "SFTPClient"]


class FTPClient(base_client.BaseClient["FTPClient"]):
    """Generic FTP client."""

    allowed_schemes = ["ftp"]

    conn: ftplib.FTP

    def __init__(self, url: Union[str, urls.URL], **kwargs) -> None:
        """Create a new ftp client based on the url starting with ftp://."""
        super().__init__(url)
        self.port = self.url.port or 21

    # Connection methods:

    def _connect(self) -> ftplib.FTP:
        logging.info(f"starting ftp connection to {self.url}")
        conn = ftplib.FTP()
        conn.connect(self.url.hostname or "", port=self.port)
        conn.login(user=self.url.username or "", passwd=self.url.password or "")
        return conn

    # Stream methods:

    @decorators.check_conn
    def get(self, writer: protocols.ByteWriter, file_path: str = None) -> None:
        """Write the contents of a remote file into the passed writer.

        Arguments:
            :file_path: The path of the remote file if not passed
                in the costructor as part of the url.
        """
        remote_path = file_path or self.url.path
        if remote_path == "":
            raise exceptions.FTPError("Missing remote file path")

        try:
            self.conn.retrbinary(f"RETR {remote_path}", writer.write)
        except Exception as e:
            raise exceptions.FTPError("Error from ftp server:" + str(e))

    @decorators.check_conn
    def put(self, reader: protocols.ByteReader, file_path: Optional[str] = None) -> None:
        """Write the contents of the reader into the remote file.

        Arguments:
            :file_path: The path of the remote file if not passed
                in the costructor as part of the url.
        """
        remote_path = file_path or self.url.path
        # storebinary only works with io.BytesIO
        buff = io.BytesIO(bytearray(reader.read()))
        self.conn.storbinary(f"STOR {remote_path}", buff)

    # Helpers:

    def _scan_mlds(self, base_url):
        entries = []
        for mlsd_entry in self.conn.mlsd(self.url.path, facts=["type"]):
            # https://docs.python.org/3/library/ftplib.html#ftplib.FTP.mlsd
            file_name = mlsd_entry[0]
            entry_type = mlsd_entry[1]["type"]

            url = urls.URL(base_url + file_name)
            if entry_type == "dir":
                entries.append(fs.build_folder_entry(url))
            else:
                entries.append(fs.build_file_entry(url))

        return entries

    def _scan_dir(self, base_url):
        """Fallback if no mlst is implemented in the server.

        Do not really hope this will work everywhere.
        """
        entries = []

        def parser(line):
            nonlocal entries
            parts = line.split()
            file_name = parts[-1]
            url = urls.URL(base_url + file_name)
            if parts[0][0] == "d":
                # first line would look drwxrwx---
                # if it's  dir
                entries.append(fs.build_folder_entry(url))
            else:
                entries.append(fs.build_file_entry(url))

        self.conn.dir(self.url.path.lstrip("/"), parser)

        return entries

    def scandir(self, **kwargs) -> Iterable[fs.DirEntry]:
        """Scan the connection url to create dir entries."""
        base_url = f"ftp://{self.url.hostname}:{self.port}{self.url.path.rstrip('/')}/"
        try:
            return self._scan_mlds(base_url)
        except ftplib.error_perm as e:
            if "501 'MLST type;'" in str(e):
                # if mlst is not implemented in the server
                # try to use dir and parse the output
                return self._scan_dir(base_url)
            else:
                raise exceptions.FTPError("Error from ftp server:" + str(e))

    @decorators.check_conn
    def remove(self):
        """Remove the file from the ftp."""
        self.conn.delete(self.url.path)


class SFTPClient(base_client.BaseClient["SFTPClient"]):
    """SFTP stream client."""

    allowed_schemes = ["sftp"]

    conn: pysftp.Connection
    username: str
    password: str
    port: int

    def __init__(self, url: Union[str, urls.URL], **kwargs) -> None:
        """Create a new sftp client based on the url starting with sftp://."""
        super().__init__(url)
        self.username = self.url.username or ""
        self.password = self.url.password or ""
        self.port = self.url.port or 22

    # Connection methods:

    def _connect(self) -> pysftp.Connection:
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        return pysftp.Connection(
            self.url.hostname,
            username=self.username,
            password=self.password,
            cnopts=cnopts,
            port=self.port,
        )

    # Stream methods:

    @decorators.check_conn
    def get(self, writer: protocols.ByteWriter, file_path: str = None) -> None:
        """Write the contents of a remote file into the passed writer.

        Arguments:
            :file_path: The path of the remote file if not passed
                in the costructor as part of the url.
        """
        remote_path = file_path or self.url.path
        if remote_path == "":
            raise exceptions.FTPError("Missing remote file path")

        if not self.conn.isfile(remote_path):
            raise exceptions.FTPError("Unable to fetch the remote file")

        logger.info(f"sftp reading from {remote_path}")
        self.conn.getfo(remote_path, writer)

    @decorators.check_conn
    def put(self, reader: protocols.ByteReader, file_path: str = None) -> None:
        """Write the contents of the reader into the remote file.

        Arguments:
            :file_path: The path of the remote file if not passed
                in the costructor as part of the url.
        """
        remote_path = file_path or self.url.path
        if remote_path == "":
            raise exceptions.FTPError("Missing remote file path")

        logger.info(f"sftp writing to {remote_path}")
        # this gives permission error
        # self.conn.putfo(remote_path, file_obj.read())
        # but open works
        with self.conn.open(remote_path, mode="wb") as f:
            f.write(reader.read())

    def scandir(self, **kwargs) -> Iterable[fs.DirEntry]:
        """Scan the connection url to create dir entries."""
        base_url = f"sftp://{self.url.hostname}:{self.port}{self.url.path}/"
        entries = []
        for attrs in self.conn.listdir_attr(self.url.path):
            url = urls.URL(base_url + attrs.filename)
            if stat.S_ISDIR(attrs.st_mode):
                entries.append(fs.build_folder_entry(url))

            elif stat.S_ISREG(attrs.st_mode):
                entries.append(fs.build_file_entry(url))
            else:
                continue  # ignore other type of entries
        return entries

    @decorators.check_conn
    def remove(self):
        """Remove the file from the ftp."""
        self.conn.remove(self.url.path)



"""General note on streams and writers/readers.

We are using byte streams (writers and readers that only operate with byte arrays)
as canonical format to avoid a huge mess in string vs bytes duality.

We still allow client code open/write string data if they want to.
But that the actual data is always send to the under lying client (ftp, s3...) as bytes.

One of the reasons is that S3 only supports streams as binary data, which is not fully pythonic
(albeit being probably more sane).

In order to offer client code the possibility to write csv/parquet files into s3 buckets without
having to worry about the underlying implementations. We hide which stream flavour (string, bytes)
is used and only exposes the bytes versions to the S3Client.

This is down by using TextIOWrapper when text is needed by the client code and exposes the inner
buffer to the underlying client.
"""
import abc
import io
from typing import IO, Any, ContextManager, Optional, Protocol

from tentaclio import protocols


class Streamer(Protocol):
    """Interface for stream-based connections."""

    # Stream methods:

    @abc.abstractmethod
    def get(self, writer: protocols.ByteWriter, **params) -> None:
        """Read the contents from the stream and write them the the ByteWriter."""
        ...

    @abc.abstractmethod
    def put(self, reader: protocols.ByteReader, **params) -> None:
        """Write the contents of the reader into the client stream."""
        ...


class StreamerContextManager(Streamer, ContextManager, Protocol):
    """Interface for stream-based connections within context managers.

    Useful as mypy doesn't recognise StreamerContext[Streamer]
    """

    ...


class StreamBaseIO:
    """Base class for IO streams that interact with Streamers.

    The extra methods included are to ensure interoperability with loosely defined,
    third party libraries such as `pyarrow`.

    https://docs.python.org/3/library/io.html
    """

    buffer: IO

    def __init__(self, buffer: IO):
        """Create a StreamBase which wraps the provided buffer."""
        self.buffer = buffer

    @property
    def closed(self):
        """Tell if the resource is closed."""
        self.buffer.closed

    def __iter__(self):
        """Return the resource as an iterable.

        For pandas compatibility
        https://github.com/pandas-dev/pandas/blob/master/pandas/core/dtypes/inference.py#L194
        """
        # for pandas compatibility
        return self.buffer.__iter__()

    def seek(self, *args, **kargs):
        """Change the stream position to the given byte offset."""
        return self.buffer.seek(*args, **kargs)

    def tell(self) -> int:
        """Return the current stream position."""
        return self.buffer.tell()

    def truncate(self, size: Optional[int] = None):
        """Resize the stream to the given size in bytes."""
        return self.buffer.truncate(size)

    def seekable(self) -> bool:
        """Mark this stream as seekable."""
        return True

    def flush(self):
        """Flush stream, to be overriden if necessary."""
        ...


class StreamerWriter(StreamBaseIO):
    """Offer stream like access to underlying client.

    Offer a IO stream interface to clients while maintaining the
    connection atomic.
    """

    def __init__(self, client: StreamerContextManager, buffer: IO):
        """Create a new writer based on a stream client and a buffer."""
        super().__init__(buffer)
        self.client = client

    def write(self, contents: Any) -> int:
        """Write the contents to the underlying buffer."""
        return self.buffer.write(contents)

    def close(self) -> None:
        """Flush and close the writer."""
        self._flush()
        self.buffer.close()

    def _flush(self):
        self.buffer.seek(0)
        # atomic put so we open/close connections swiftly
        with self.client:
            self.client.put(self.buffer)


class StreamerReader(StreamBaseIO):
    """Offer stream like access to underlying client.

    Offer a IO stream interface to clients while maintaining the
    connection atomic.
    """

    buffer: IO

    def __init__(self, client: StreamerContextManager, buffer: IO):
        """Create a reader that will read from the given client to the passed buffer."""
        super().__init__(buffer)
        self.client = client
        self._load()

    def _load(self):
        # atomic get so we open/close connections swiftly
        with self.client:
            self.client.get(self.buffer)
        self.buffer.seek(0)

    def read(self, size: int = -1):
        """Read the contents of the buffer."""
        return self.buffer.read(size)

    def readline(self, size: int = -1):
        """Read and return one line from the buffer."""
        # Additional method required for unpickling Python objects
        return self.buffer.readline(size)

    def close(self) -> None:
        """Close the writer."""
        self.buffer.close()


class StringToBytesClientReader(StreamerReader):
    """String based stream reader that uses a byte buffer under the hood.

    This is used to allow clients (s3,ftp...) just use bytes while the code
    using this reader read data as strings.
    This is similar to how python itself treats text/binary files.
    """

    inner_buffer: io.BytesIO

    def __init__(self, client: StreamerContextManager):
        """Create a byte based reader that will read from the given client."""
        self.inner_buffer = io.BytesIO()
        super().__init__(client, io.TextIOWrapper(self.inner_buffer, encoding="utf-8"))

    def _load(self):
        # interacts with the client in terms of bytes
        with self.client:
            self.client.get(self.inner_buffer)
        self.buffer.seek(0)


class StringToBytesClientWriter(StreamerWriter):
    """String based stream reader that uses a byte buffer under the hood.

    This is used to allow clients (s3,ftp...) just use bytes while the code
    using this writer can write data as strings.
    This is similar to how python itself treats text/binary files.
    """

    inner_buffer: io.BytesIO

    def __init__(self, client: StreamerContextManager):
        """Create a byte based write that will read from the given client."""
        self.inner_buffer = io.BytesIO()
        super().__init__(client, io.TextIOWrapper(self.inner_buffer, encoding="utf-8"))

    def _flush(self) -> None:
        """Flush and close the writer."""
        self.buffer.seek(0)
        # interacts with the client in terms of bytes
        with self.client as client:
            client.put(self.inner_buffer)



"""Method for importing tentaclio plugins."""
import importlib
import logging
from typing import Callable, Iterable

from importlib_metadata import packages_distributions


PackageLister = Callable[[], Iterable[str]]

logger = logging.getLogger(__name__)


def import_tentaclio_plugins(
    list_packages: PackageLister = packages_distributions().keys,
) -> None:
    """Find and import tentaclio plugins."""
    for package in list_packages():
        if package.startswith("tentaclio_"):
            logger.info(f"Importing plugin: {package}")
            importlib.import_module(package)



"""Module to give support to miscellaneous fs like operations over urls."""

from .copier import *  # noqa
from .copiers import *  # noqa
from .remover import *  # noqa
from .scanner import *  # noqa
from .scanners import *  # noqa



"""Main entry points to tentaclio-io."""
from typing import ContextManager

from tentaclio import protocols
from tentaclio.credentials import authenticate

from .stream_registry import STREAM_HANDLER_REGISTRY


__all__ = ["open"]

VALID_MODES = ("", "rb", "wb", "rt", "wt", "r", "w", "b", "t")


def open(url: str, mode: str = None, **kwargs) -> ContextManager[protocols.AnyReaderWriter]:
    """Open a url and return a reader or writer depending on mode.

    Arguments:
        :mode: similar to built-in open, allowed modes are combinations of "w" for writing
        "r" for reading. "t" for text resources, and "b" for binary. The default is "rt".
    Examples:
        >>> open(path, 'b')  # opens binary reader
        >>> open(path, 'br')  # opens binary reader
        >>> open(path, 'wb')  # opens binary writer
        >>> open(path, 'wt')  # opens text  writer

    """
    mode = mode or ""

    _assert_mode(mode)
    is_write_mode = "w" in mode

    if is_write_mode:
        return _open_writer(url=url, mode=mode, **kwargs)
    else:
        return _open_reader(url=url, mode=mode, **kwargs)


# Helpers


def _assert_mode(mode: str):
    """Check if a mode is valid or raise an error otherwise."""
    if mode not in VALID_MODES:
        valid_modes = ",".join(VALID_MODES)
        raise ValueError(f"Mode {mode} is not allowed. Valid modes are  {valid_modes}")


def _open_writer(url: str, mode: str, **kwargs) -> ContextManager[protocols.Writer]:
    """Open a url and return a writer."""
    authenticated = authenticate(url)
    return STREAM_HANDLER_REGISTRY.open_stream_writer(authenticated, mode, extras=kwargs)


def _open_reader(url: str, mode: str, **kwargs) -> ContextManager[protocols.Reader]:
    """Open a url and return a reader."""
    authenticated = authenticate(url)
    return STREAM_HANDLER_REGISTRY.open_stream_reader(authenticated, mode, extras=kwargs)



"""Stream handler registry to open readers and writers to urls."""
from typing import ClassVar, ContextManager, Optional, Protocol

from tentaclio import protocols
from tentaclio.registry import URLHandlerRegistry
from tentaclio.urls import URL


__all__ = ["STREAM_HANDLER_REGISTRY"]


class _ReaderContextManager(ContextManager[protocols.Reader]):
    """Composed context manager for ReaderCloseables."""

    def __init__(self, resource: protocols.ReaderClosable):
        super(_ReaderContextManager, self).__init__()
        self.resource = resource

    def __enter__(self) -> protocols.Reader:
        return self.resource

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return self.resource.close()


class _WriterContextManager(ContextManager[protocols.Writer]):
    """Composed context manager for WriterCloseables."""

    def __init__(self, resource: protocols.WriterClosable):
        super(_WriterContextManager, self).__init__()
        self.resource = resource

    def __enter__(self) -> protocols.Writer:
        return self.resource

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        return self.resource.close()


class StreamHandler(Protocol):
    """Protocol for handling stream resources."""

    def open_reader_for(self, url: "URL", mode: str, extras: dict) -> protocols.ReaderClosable:
        """Open a reader for the given url."""
        ...

    def open_writer_for(self, url: "URL", mode: str, extras: dict) -> protocols.WriterClosable:
        """Open a writer for the given url."""
        ...


class StreamHandlerRegistry(URLHandlerRegistry[StreamHandler]):
    """Registry for stream handlers."""

    def open_stream_reader(
        self, url: URL, mode: str, extras: Optional[dict] = None
    ) -> ContextManager[protocols.Reader]:
        """Open a reader for the stream located at this url."""
        extras = extras or {}
        reader = self.get_handler(url.scheme).open_reader_for(url, mode, extras)
        return _ReaderContextManager(reader)

    def open_stream_writer(
        self, url: URL, mode: str, extras: Optional[dict] = None
    ) -> ContextManager[protocols.Writer]:
        """Open a writer for the stream located at this url."""
        extras = extras or {}
        writer = self.get_handler(url.scheme).open_writer_for(url, mode, extras)
        return _WriterContextManager(writer)


class _StreamRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[StreamHandlerRegistry] = StreamHandlerRegistry()


STREAM_HANDLER_REGISTRY = _StreamRegistryHolder().instance



"""Streams hide clients and allow access to stream resources via a simple read/write interface."""
from .api import *  # noqa
from .stream_client_handler import *  # noqa
from .stream_registry import *  # noqa



"""tentaclio is a library that tries to unify how to load, store data.

The main use cases in mind are ETL processes and notebooks, but can be used in many other contexts.
The main benefits are:
    * url based resource management.
    * same function to open readers and writers for resources of different natures.
    Just change the url, the code remains the same.
    * The same for dbs, create clients with ease and use them regardless
    the underlying implementation (thanks to sqlalchemy).
    * Credentials management that allows a distributed credentials storage.
"""
from .clients import *  # noqa
from .databases import *  # noqa
from .databases.api import *  # noqa
from .fs import *  # noqa
from .fs.api import *  # noqa
from .protocols import *  # noqa
from .streams import *  # noqa
from .urls import *  # noqa


import_tentaclio_plugins()

# Stream handlers

# Local files
STREAM_HANDLER_REGISTRY.register("", StreamURLHandler(LocalFSClient))
STREAM_HANDLER_REGISTRY.register("file", StreamURLHandler(LocalFSClient))

# ftp / sftp handlers
STREAM_HANDLER_REGISTRY.register("ftp", StreamURLHandler(FTPClient))
STREAM_HANDLER_REGISTRY.register("sftp", StreamURLHandler(SFTPClient))

# http / https handlers
STREAM_HANDLER_REGISTRY.register("http", StreamURLHandler(HTTPClient))
STREAM_HANDLER_REGISTRY.register("https", StreamURLHandler(HTTPClient))

# Directory Scanners

SCANNER_REGISTRY.register("", ClientDirScanner(LocalFSClient))
SCANNER_REGISTRY.register("file", ClientDirScanner(LocalFSClient))


SCANNER_REGISTRY.register("ftp", ClientDirScanner(FTPClient))
SCANNER_REGISTRY.register("sftp", ClientDirScanner(SFTPClient))


REMOVER_REGISTRY.register("", ClientRemover(LocalFSClient))
REMOVER_REGISTRY.register("file", ClientRemover(LocalFSClient))
REMOVER_REGISTRY.register("ftp", ClientRemover(FTPClient))
REMOVER_REGISTRY.register("sftp", ClientRemover(SFTPClient))



"""Functionality for listing directory-like urls."""
from typing import ClassVar, Iterable, Protocol

from tentaclio.registry import URLHandlerRegistry
from tentaclio.urls import URL


__all__ = ["SCANNER_REGISTRY", "DirEntry", "build_file_entry", "build_folder_entry"]


class DirEntry:
    """Entry containing information about a directory scan item."""

    def __init__(self, *, url: URL, is_dir: bool, is_file: bool):
        """Create new dir entry."""
        self.url: URL = url
        self.is_dir: bool = is_dir
        self.is_file: bool = is_file


def build_folder_entry(url: URL) -> DirEntry:
    """Build a DirEntry with is_dir = True."""
    return DirEntry(url=url, is_file=False, is_dir=True)


def build_file_entry(url: URL) -> DirEntry:
    """Build a DirEntry with is_file = True."""
    return DirEntry(url=url, is_file=True, is_dir=False)


class Scanner(Protocol):
    """Scan a directory-like url."""

    def scandir(self, url: URL) -> Iterable[DirEntry]:
        """Scan a directory-like url returning its entries.

        Return an iterator of tentaclio.fs.DirEntry objects corresponding
        to the entries in the directory given by url. The entries
        are yielded in arbitrary order,
        and the special entries '.' and '..' are not included.

        More info:
        https://docs.python.org/3/library/os.html#os.scandir
        """
        ...


class ScannerRegistry(URLHandlerRegistry[Scanner]):
    """Registry for scanners."""

    ...


class _ScannerRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[ScannerRegistry] = ScannerRegistry()


SCANNER_REGISTRY = _ScannerRegistryHolder().instance



"""URL related clasess."""
import logging
from typing import Any, Dict, Optional
from urllib import parse


__all__ = ["URL"]

logger = logging.getLogger(__name__)


class URLError(Exception):
    """Error encountered while processing a URL."""


def _netloc_from_components(username=None, password=None, hostname=None, port=None):
    """Construct network location in accordance with urllib."""
    netloc = ""
    if username is not None:
        netloc += parse.quote(username, safe="")
        if password is not None:
            netloc += ":" + parse.quote(password, safe="")
        netloc += "@"
    hostname = hostname or ""
    netloc += hostname
    if port is not None:
        netloc += ":" + str(port)
    return netloc


class URL:
    """Immutable class that give access to single url components and allows access to streams."""

    _scheme: str
    _username: Optional[str] = None
    _password: Optional[str] = None
    _hostname: Optional[str] = None
    _port: Optional[int] = None
    _path: str
    _query: Optional[Dict[str, str]] = None

    def __init__(self, url: str) -> None:
        """Create a url by parsing the parametre."""
        self._url = url
        self._parse_url()

    # Helpers:
    def _parse_url(self) -> None:
        if self._url is None:
            raise URLError("None url")

        parsed_url = parse.urlparse(self._url)

        self._scheme = parsed_url.scheme
        self._username = parsed_url.username
        self._password = parsed_url.password
        self._hostname = parsed_url.hostname
        self._port = parsed_url.port
        self._path = parsed_url.path

        # Replace %xx escapes - ONLY for username & password
        if parsed_url.username and self._username:
            self._username = parse.unquote(self._username)
        if parsed_url.password and self._password:
            self._password = parse.unquote(self._password)

        if parsed_url.query:
            # Assuming string values
            self._query = {
                key: value for key, value in parse.parse_qsl(parsed_url.query, strict_parsing=True)
            }
        else:
            self._query = None

    def copy(
        self,
        scheme: str = None,
        username: str = None,
        password: str = None,
        hostname: str = None,
        port: int = None,
        path: str = None,
        query: str = None,
    ) -> "URL":
        """Copy this url optionally overwriting the provided components."""
        return URL.from_components(
            scheme=scheme or self.scheme,
            username=username or self.username,
            password=password or self.password,
            hostname=hostname or self.hostname,
            port=port or self.port,
            path=path or self.path,
            query=query or self.query,
        )

    @classmethod
    def from_components(
        cls,
        scheme=None,
        username=None,
        password=None,
        hostname=None,
        port=None,
        path=None,
        query=None,
    ):
        """Construct a URL object from parts."""
        netloc = _netloc_from_components(
            username=username, password=password, hostname=hostname, port=port
        )
        params = None
        if query:
            query = parse.urlencode(query)
        fragment = None
        components = (scheme, netloc, path, params, query, fragment)
        url = parse.urlunparse(components)
        return URL(url)

    def __str__(self):
        """Return a string representation of the url hidding compromising credentials."""
        password = self.password
        if password:
            password = "__secret__" + password[-4:]
        return self.copy(password=password)._url

    def __eq__(self, other: Any):
        """Check if two urls are equal."""
        if not isinstance(other, URL):
            return False
        return self._url == other._url

    def __repr__(self):
        """Return the string url representation."""
        return f"URL({self._url})"

    @property
    def scheme(self) -> str:
        """Access the scheme."""
        return self._scheme

    @property
    def username(self) -> Optional[str]:
        """Access the username."""
        return self._username

    @property
    def password(self) -> Optional[str]:
        """Access the password."""
        return self._password

    @property
    def hostname(self) -> Optional[str]:
        """Access the hostname."""
        return self._hostname

    @property
    def port(self) -> Optional[int]:
        """Access the port."""
        return self._port

    @property
    def path(self) -> str:
        """Access the path."""
        return self._path

    @property
    def query(self) -> Optional[Dict[str, str]]:
        """Access the query."""
        return self._query

    @property
    def url(self) -> str:
        """Return the original url."""
        return self._url



"""HTTP Stream client."""
import io
from typing import Optional, Union
from urllib import parse

import requests

from tentaclio import protocols

from . import base_client, decorators, exceptions


__all__ = ["HTTPClient"]


DEFAULT_TIMEOUT = 10.0

DEFAULT_HEADERS = {"Accept": "application/json"}


class HTTPClient(base_client.BaseClient["HTTPClient"]):
    """HTTP stream client.

    This client is useful when dealing inputs that may change from local files to http,
    or from s3 to http in simple usecases. We're not intending to wrap requests or rewrite it.
    """

    allowed_schemes = ["http", "https"]

    conn: requests.Session
    timeout: float
    headers: dict
    protocol: str
    username: Optional[str]
    password: Optional[str]
    hostname: str
    port: Optional[int]
    endpoint: str

    def __init__(
        self, url: str, default_timeout: float = None, default_headers: dict = None
    ) -> None:
        """Create a new http/https client based on the passed url and extra params."""
        # Default connection timeout at 10''
        self.timeout = default_timeout or DEFAULT_TIMEOUT
        # Default JSON response back
        self.headers = default_headers or DEFAULT_HEADERS
        super().__init__(url)

        self.protocol = self.url.scheme

        if self.url.hostname is None:
            raise exceptions.HTTPError("Missing URL hostname")
        self.hostname = self.url.hostname

        self.port = self.url.port
        self.endpoint = self.url.path

        # Enforce no empty credentials
        self.username = None if self.url.username == "" else self.url.username
        self.password = None if self.url.password == "" else self.url.password

    # Connection methods:

    def _connect(self) -> requests.Session:
        session = requests.Session()

        # credentials provided
        if self.username and self.password:
            session.auth = (self.username, self.password)

        # Non-empty header
        if self.headers:
            session.headers.update(self.headers)

        return session

    # Stream methods:

    @decorators.check_conn
    def get(
        self,
        writer: protocols.ByteWriter,
        endpoint: str = None,
        params: dict = None,
        options: dict = None,
    ) -> None:
        """Read the contents from the url and write them into the provided writer.

        Arguments:
            :end_point: Path to append to the url passed in the constructor.
            :params: Url params to add
            :options: More options for the request library.
        """
        url = self._fetch_url(endpoint or "")

        request = self._build_request("GET", url, default_params=params)
        response = self._send_request(request, default_options=options)

        writer.write(response.content)

    @decorators.check_conn
    def put(
        self,
        reader: protocols.ByteReader,
        endpoint: str = None,
        params: dict = None,
        options: dict = None,
    ) -> None:
        """Write the contents of the provided reader into the url using POST.

        Arguments:
            :end_point: Path to append to the url passed in the constructor.
            :params: Url params to add
            :options: More options for the request library.
        """
        url = self._fetch_url(endpoint or "")
        buff = io.StringIO(bytes(reader.read()).decode(encoding="utf-8"))
        request = self._build_request("POST", url, default_data=buff, default_params=params)
        self._send_request(request, default_options=options)

    # Helpers:

    def _fetch_url(self, endpoint: str) -> str:
        if endpoint == "" and self.endpoint == "":
            raise exceptions.HTTPError("Missing URL end point")
        # Fetch full base URL
        base_url = parse.urlunparse((self.protocol, self.hostname, self.endpoint, "", "", ""))
        return parse.urljoin(base_url, endpoint)

    def _build_request(
        self,
        method: str,
        url: str,
        default_data: protocols.Reader = None,
        default_params: dict = None,
    ):
        data: Union[protocols.Reader, list] = default_data or []
        params = default_params or {}

        if method == "GET":
            # GET uses params
            request = requests.Request(method, url, params=params, headers=self.headers)
        elif method == "POST":
            # POST uses data & params
            request = requests.Request(method, url, data=data, params=params, headers=self.headers)
        else:
            raise NotImplementedError

        return self.conn.prepare_request(request)

    def _send_request(self, request: requests.PreparedRequest, default_options: dict = None):
        options = default_options or {}

        response = self.conn.send(
            request,
            stream=options.get("stream", False),
            verify=options.get("verify", False),
            proxies=options.get("proxies", {}),
            cert=options.get("cert"),
            timeout=options.get("timeout", self.timeout),
            allow_redirects=options.get("allow_redirects", True),
        )

        if options.get("check_response", True):
            self._check_response(response)

        return response

    @staticmethod
    def _check_response(response: requests.Response) -> None:
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            raise exceptions.HTTPError(f"{response.status_code}: {response.reason}")



"""Helper cli to encode urls."""
import typing

import click

from tentaclio import urls


SCOPES = ["https://www.googleapis.com/auth/drive"]


# Register CLI commands
@click.group()
def main():
    """Run tentaclio helper commands."""
    ...


@main.group()
def url():
    """Run url related helpers."""
    ...


@url.command()
@click.option("--scheme", default=None, required=True)
@click.option("--username", default=None)
@click.option("--password", default=None)
@click.option("--hostname", default=None)
@click.option("--port", default=None, type=int)
@click.option("--path", default=None)
@click.option(
    "--key",
    default=None,
    nargs=2,
    multiple=True,
    help="Provide key-value pairs using [--key KEY VALUE]",
)
def compose_url(scheme, username, password, hostname, port, path, key):
    """Compose a client URL from individual components."""
    if not hostname and not path:
        raise click.ClickException("Provide at least one of `hostname`, `path`")

    query: typing.Optional[dict] = None
    if key:
        query = {k: v for k, v in key}

    composed_url = urls.URL.from_components(
        scheme=scheme,
        username=username,
        password=password,
        hostname=hostname,
        port=port,
        path=path,
        query=query,
    )

    # Output composed url (with no hidden secrets)
    click.echo(composed_url.url, nl=False)


if __name__ == "__main__":
    main(prog_name="python -m tentaclio")



"""Client exceptions."""


class ClientError(Exception):
    """Basic exception raised for a given product."""


class FTPError(ClientError):
    """Exception encountered over a FTP client connection."""


class HTTPError(ClientError):
    """Exception encountered over a HTTP client connection."""



"""Csv to database stream like access."""
import csv
import io
from typing import IO, ContextManager, Protocol, Sequence

from tentaclio import protocols

from . import base_stream


class CsvDumper(ContextManager, Protocol):
    """Csv into db dumping contract."""

    def dump_csv(
        self, csv_reader: protocols.Reader, columns: Sequence[str], dest_table: str
    ) -> None:
        """Dump the csv contents into the the given table."""
        ...


def _get_field_names(reader: IO) -> Sequence[str]:
    """Get the field names from the contents."""
    csv_reader = csv.DictReader(reader)
    next(csv_reader)
    reader.seek(0)
    return csv_reader.fieldnames or []


class DatabaseCsvWriter(base_stream.StreamBaseIO):
    """Writer that dumps the csv formatted data into the specified table.

    The connection is done through the relevant client.
    The csv data is expected to be native's python dialect. namely first row defines the header and
    the data is delimited by `,`.
    """

    def __init__(self, csv_dumper: CsvDumper, table: str) -> None:
        """Create a new csv writer initialising the internal buffer.

        :csv_dumper: an object that is able to write csv files into a table.
        :table: the name of the destination table in the database.
        """
        self.csv_dumper = csv_dumper
        self.table = table
        super().__init__(io.StringIO())

    def write(self, contents) -> int:
        """Append contents to the internal buffer."""
        return self.buffer.write(contents)

    def _flush(self) -> None:
        """Use the client to dump the data into the configured table."""
        self.buffer.seek(0)
        field_names = _get_field_names(self.buffer)
        with self.csv_dumper:
            self.csv_dumper.dump_csv(self.buffer, field_names, self.table)

    def close(self) -> None:
        """Close the internal buffer and flush the contents to db."""
        self._flush()
        self.buffer.close()

    def __enter__(self) -> protocols.Writer:
        """Return self as Writer."""
        return self

    def __exit__(self, *args) -> None:
        """Close the writer."""
        self.close()



"""Read credentials from yaml file."""
import io
import logging
import os

import yaml

from tentaclio import protocols, urls
from tentaclio.credentials import injection


logger = logging.getLogger(__name__)

SECRETS = "secrets"
TENTACLIO_SECRETS_FILE = "TENTACLIO__SECRETS_FILE"


def add_credentials_from_env_file(
    injector: injection.CredentialsInjector,
) -> injection.CredentialsInjector:
    """Add credentials from a configurable yaml file.

    The path of this file is expected to be set in the environment as
    `TENTACLIO__SECRETS_FILE`
    """
    secrets_file = os.getenv(TENTACLIO_SECRETS_FILE, "")
    if secrets_file != "":
        injector = _load_from_file(injector, secrets_file)
    return injector


def _load_creds_from_yaml(yaml_reader: protocols.Reader) -> dict:
    loaded_data = yaml.safe_load(io.StringIO(yaml_reader.read()))
    if SECRETS not in loaded_data:
        raise KeyError(f"no secrets in yaml data. Make sure the file has a '{SECRETS}' element")
    return loaded_data[SECRETS]


def _load_from_file(
    injector: injection.CredentialsInjector, path: str
) -> injection.CredentialsInjector:
    try:
        with open(path, "r") as f:
            return add_credentials_from_reader(injector, f)
    except IOError as e:
        logger.error("Error while loading secrets file {path}")
        raise e


def add_credentials_from_reader(
    injector: injection.CredentialsInjector, yaml_reader: protocols.Reader
) -> injection.CredentialsInjector:
    """Read the credentials from a yml.

    The file has the follwing format:
        secrets:
            my_creds_name: http://user:password@google.com/path
            my_db: postgres://user_db:password@octoenergy.com/databasek

    """
    creds = _load_creds_from_yaml(yaml_reader)
    for name, url in creds.items():
        logger.info(f"Adding secret: {name}")
        try:
            injector.register_credentials(urls.URL(url))
        except Exception as e:
            logger.error(f"Error while registering credentials {name}:{url} from file")
            raise e

    return injector



"""Load credentials from the environment."""
import logging
import os
from typing import Dict, List, Optional, cast

from tentaclio import urls

from .injection import CredentialsInjector


logger = logging.getLogger(__name__)


Env = Dict[str, str]

TENTACLIO_CONN_PREFIX = "TENTACLIO__CONN__"


def _get_connection_urls(env: Env) -> List[urls.URL]:
    connections = []
    for key, val in env.items():
        if key.startswith(TENTACLIO_CONN_PREFIX):
            try:
                connections.append(urls.URL(val))
            except Exception as e:
                logger.error(f"Error parsing credentials from the environment variable {key}")
                raise e
    return connections


def add_credentials_from_env(
    injector: CredentialsInjector, env: Optional[Env] = None
) -> CredentialsInjector:
    """Add urls with credentials from the environment.

    this funciton will scan the environment and add the variables
    with the prefix TENTACLIO_CONN__ to the pool of
    available connection strings.
    """
    if env is None:
        env = cast(Env, os.environ)
    urls = _get_connection_urls(env)
    for url in urls:
        injector.register_credentials(url)
    return injector



"""Provide sql connection using sqlalchemy.

This client is used for convinience when using different sql
providers and unifying the client creation. We do not intent to rewriter sqlalchemy.
"""
import contextlib
from typing import Container, Generator, Optional, Union

import pandas as pd
from sqlalchemy.engine import Connection, Engine, create_engine, result
from sqlalchemy.engine import url as sqla_url
from sqlalchemy.orm import session, sessionmaker
from sqlalchemy.sql.schema import MetaData

from tentaclio import urls

from . import base_client, decorators


__all__ = ["SQLAlchemyClient", "bound_session", "atomic_session"]


SessionGenerator = Generator[None, session.Session, None]


class _TrueContainer(Container[str]):
    """String container that always returns true.

    As we don't have control over the protocols that sqlalchemy is able to
    accept. We shouldn't try to limit which urls can be used here or not.
    So this container will play well with the super class checks to allow
    schemes for different clients.
    """

    def __contains__(self, obj: object) -> bool:
        """Return true."""
        return True


class SQLAlchemyClient(base_client.BaseClient["SQLAlchemyClient"]):
    """SQLAlchemy based client."""

    # The allowed drivers depend on the dependencies installed.
    allowed_schemes: Container[str] = _TrueContainer()
    # Default connect_args
    connect_args_default: dict = {}

    conn: Connection
    engine: Optional[Engine] = None
    execution_options: dict
    connect_args: dict
    database: str
    drivername: str
    username: Optional[str]
    password: Optional[str]
    host: Optional[str]
    port: Optional[int]

    def __init__(
        self, url: Union[str, urls.URL], execution_options: dict = None, connect_args: dict = None
    ) -> None:
        """Create sqlalchemy client based on the passed url.

        This is a wrapper for sqlalchemy engine/connection creation.
        """
        self.execution_options = execution_options or {}
        self.connect_args = connect_args or self.connect_args_default
        super().__init__(url)
        self._extract_url_params()

    def _extract_url_params(self) -> None:
        """Extract the database parameters from the url."""
        # the database doesn't start with /
        database = self.url.path[1:]

        self.database = database
        self.drivername = self.url.scheme
        self.username = self.url.username
        self.password = self.url.password
        self.host = self.url.hostname
        self.port = self.url.port
        self.url_query = self.url.query

    # Connection methods:

    def _connect(self) -> Connection:

        parsed_url = sqla_url.URL(
            drivername=self.drivername,
            username=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
            query=self.url_query,
        )
        if self.engine is None:
            self.engine = create_engine(
                parsed_url,
                execution_options=self.execution_options,
                connect_args=self.connect_args,
            )
        return self.engine.connect()

    def _get_raw_conn(self):
        """Acquire raw DBAPI connection from the pool."""
        return self.conn.engine.raw_connection()

    # Schema methods:

    def set_schema(self, meta_data: MetaData) -> None:
        """Create tables based on the metadata object."""
        meta_data.create_all(bind=self.conn)

    def delete_schema(self, meta_data: MetaData) -> None:
        """Delete tables based on the metadata object."""
        meta_data.drop_all(bind=self.conn)

    # Query methods:

    @decorators.check_conn
    def query(self, sql_query: str, **kwargs) -> result.ResultProxy:
        """Execute a read-only SQL query, and return results.

        This will not commit any changes to the database.
        """
        return self.conn.execute(sql_query, **kwargs)

    @decorators.check_conn
    def execute(self, sql_query: str, **kwargs) -> None:
        """Execute a raw SQL query command."""
        trans = self.conn.begin()
        try:
            self.conn.execute(sql_query, **kwargs)
        except Exception:
            trans.rollback()
            raise
        else:
            trans.commit()

    # Dataframe methods:

    @decorators.check_conn
    def get_df(self, sql_query: str, params: dict = None, **kwargs) -> pd.DataFrame:
        """Run a raw SQL query and return a data frame."""
        return pd.read_sql(sql_query, self.conn, params=params, **kwargs)


# Session context managers:


@contextlib.contextmanager
def bound_session(connection: Connection) -> SessionGenerator:
    """Context manager for a sqlalchemy session."""
    Session = sessionmaker()
    sess = Session(bind=connection)
    try:
        yield sess
    finally:
        sess.close()


@contextlib.contextmanager
def atomic_session(connection: Connection) -> SessionGenerator:
    """Context manager for a session that will rollback in case of an exception."""
    Session = sessionmaker()
    sess = Session(bind=connection)
    try:
        yield sess
    except Exception:
        sess.rollback()
        raise
    else:
        sess.commit()
    finally:
        sess.close()



"""Protocols used in tentaclio."""
from typing import Any, Generic, Protocol, TypeVar, Union


__all__ = [
    "Closable",
    "Reader",
    "Writer",
    "ByteReader",
    "ByteWriter",
    "ReaderClosable",
    "WriterClosable",
    "AnyReaderWriter",
]

# https://blog.daftcode.pl/covariance-contravariance-and-invariance-the-ultimate-python-guide-8fabc0c24278
# For it's used as a Generic in a protocol T_read needs to be covariant and T_Write contravariant.
T_read = TypeVar("T_read", covariant=True)
T_write = TypeVar("T_write", contravariant=True)


class Closable(Protocol):
    """Closable protocol."""

    def close(self) -> None:
        """Close the resource."""
        ...


class _Reader(Generic[T_read], Protocol):
    def read(self, size: int = -1) -> T_read:
        ...


class _Writer(Generic[T_write], Protocol):
    def write(self, contents: T_write) -> int:
        ...


class Reader(_Reader[Any], Protocol):
    """Reader protocol without buffer specialisation."""

    ...


class Writer(_Writer[Any], Protocol):
    """Writer protocol without buffer specialisation."""

    ...


class ByteReader(_Reader[bytes], Protocol):
    """Reader protocol for bytestrings."""

    ...


class ByteWriter(_Writer[bytes], Protocol):
    """Writer protocol for bytestrings."""

    ...


# Attention! Protocol needs to be the last superclass!
# otherwise returning these composed protocols won't work
class ReaderClosable(Reader, Closable, Protocol):
    """Composed protocol for reader/closeables."""

    ...


class WriterClosable(Writer, Closable, Protocol):
    """Composed protocol for reader/closeables."""

    ...


AnyReaderWriter = Union[Reader, Writer]
